import timeit

code = """
# consensus = _quickconsensus.MismatchCounts("4", 96549060, 96549060+1000)
# consensus = quickconsensus.MismatchCounts("4", 96549060, 96549060+1000)
consensus = _quickconsensus.MismatchCounts("4", 96549060, 96567077)
# consensus = quickconsensus.MismatchCounts("4", 96549060, 96567077)
consensus.tally_reads(bam)
# print(consensus.counts)
"""

setup = """
import pysam

import pyximport; pyximport.install()
import _quickconsensus

bam = pysam.AlignmentFile("../test/quick_consensus_test.bam")
"""

times = timeit.repeat(code, setup, number=1, repeat=3)

import numpy

print(f"{numpy.mean(times):.2f}+/-{numpy.std(times):.2f}s")


import pysam
from genomeview import MismatchCounts


def test_quick_consensus():
    bam = pysam.AlignmentFile("quick_consensus_test.bam")

    consensus = MismatchCounts("4", 96549060, 96567077)

    consensus.tally_reads(bam)


def test_compare_quick_consensus():
    from genomeview import quickconsensus
    from genomeview import _quickconsensus

    bam = pysam.AlignmentFile("quick_consensus_test.bam")

    cython_consensus = _quickconsensus.MismatchCounts("4", 96549060, 96549060+1000)
    python_consensus = quickconsensus.MismatchCounts("4", 96549060, 96549060+1000)

    python_consensus.tally_reads(bam)
    cython_consensus.tally_reads(bam)

    assert (cython_consensus.counts == python_consensus.counts).all()
    assert (cython_consensus.insertions == python_consensus.insertions).all()




import math

from genomeview.track import Track

class Axis(Track):
    def __init__(self, name):
        self.name = name
        self.scale = None

        self.height = 50

    def render(self, renderer):
        start, end = self.scale.start, self.scale.end
        yield from renderer.line(self.scale.topixels(start), 5, self.scale.topixels(end), 5)
        
        ticks = get_ticks(start, end, max(1,self.scale.pixel_width/120))
        for i, (tick, label) in enumerate(ticks):
            x = self.scale.topixels(tick)
            if x < 0 or x > self.scale.pixel_width: continue

            yield from renderer.line(x, 0, x, 20)
            anchor = "middle"
            if x < 50 and i == 0:
                anchor = "start"
                x = min(x, 5)
            elif x > self.scale.pixel_width-50 and i == len(ticks)-1:
                anchor = "end"
                x = max(x, self.scale.pixel_width-5)


            yield from renderer.text(x, 35, label, anchor=anchor, size=16)

def get_ticks(start, end, target_n_labels=10):
    ticks = []
    start = int(start)
    end = int(end)
    width = end - start

    res = (10 ** round(math.log10(end - start))) / (10**math.floor(math.log10(target_n_labels)))

    if width / res > target_n_labels*2:
        res *= 5
    elif width / res > target_n_labels*1.5:
         res *= 2.5
    elif width / res < target_n_labels*0.5:
        res /= 4.0
    elif width / res < target_n_labels*0.8:
        res /= 2.0

    roundStart = start - (start%res)

    for i in range(int(roundStart), end, int(res)):
        res_digits = math.log10(res)
        if res_digits >= 6:
            label = "{}mb".format(i / 1e6)
        elif res_digits >= 3:
            label = "{:,}kb".format(i / 1e3)
        else:
            label = "{:,}".format(i)

        ticks.append((i, label))

    return ticks


if __name__ == '__main__':
    params = [(10000, 11060, 10),
              (0, 10311, 10),
              (0, 143600, 10),
              (0, 1002552, 10)]

    for param in params:
        print(params)
        print(get_ticks(*param))
        print("")


import numpy

from genomeview.utilities import match_chrom_format

class MismatchCounts(object):
    """
    keeps track of how many of each nucleotide (or insertion/deletion) are present at each position
    -> used for the quick consensus mode
    """
    types_to_id = {"A":0, "C":1, "G":2, "T":3, "DEL":5}

    def __init__(self, chrom, start, end):
        self.chrom = chrom
        self.start = start
        self.end = end

        length = end - start
        self.counts = numpy.zeros([6, length])#, dtype="uint8")
        self.insertions = numpy.zeros(length)#, dtype="uint8")

    def tally_reads(self, bam):
        depths = []
        for pileupcolumn in bam.pileup(match_chrom_format(self.chrom, bam.references), self.start, self.end, truncate=True):
            depths.append(pileupcolumn.n)
            for pileupread in pileupcolumn.pileups:
                if pileupread.is_refskip:
                    continue
                elif pileupread.is_del:
                    self.add_count(pileupcolumn.pos, "DEL")
                else:
                    nuc =  pileupread.alignment.query_sequence[pileupread.query_position]
                    if nuc != "N":
                        self.add_count(pileupcolumn.pos, nuc)
                if pileupread.indel > 0:
                    self.add_count(pileupcolumn.pos, "INS")

    def add_count(self, position, type_):
        position -= self.start
        if type_ == "INS":
            self.insertions[position] += 1
        else:
            row = self.types_to_id[type_]
            self.counts[row,position] += 1

    # def counts(self, position):
    #     position -= self.start
    #     return self.counts[position,:]

    def query(self, type_, start, end=None):
        if start < self.start or start >= self.end:
            return False

        start -= self.start
        if end is None:
            end = start
        else:
            end -= self.start

        total = self.counts[:,start:(end+1)].sum(axis=0)
        total = total.astype(float)

        if (type_ == "INS"):
            insertions = self.insertions[start:(end+1)]
            if (insertions.sum() / total.sum()) > 0.2:
               return True

        else:        
            row = self.types_to_id[type_]
            this_type = self.counts[row,start:(end+1)]

            if type_ == "DEL":
                if  ((this_type / total) > 0.3).any():
                    return True
            elif ((this_type / total) > 0.2).any():
                return True

        return False



#!/usr/bin/env python

from setuptools import setup, find_packages, Extension

def get_version(string):
    """ Parse the version number variable __version__ from a script. """
    import re
    version_re = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version_str = re.search(version_re, string, re.M).group(1)
    return version_str


def get_includes():
    class Includes:
        def __iter__(self):
            import pysam
            import numpy
            return iter(pysam.get_include()+[numpy.get_include()])
        def __getitem__(self, i):
            return list(self)[i]
    return Includes()

def get_defines():
    class Defines:
        def __iter__(self):
            import pysam
            return iter(pysam.get_defines())
        def __getitem__(self, i):
            return list(self)[i]
    return Defines()


setup(name="genomeview",
      version=get_version(open('genomeview/__init__.py').read()),
      description="genomeview",
      author="Noah Spies",
      packages=find_packages(),

      setup_requires=["cython"],
      install_requires=["pysam", "numpy"], 
      ext_modules = [
          Extension("genomeview._quickconsensus",
                  sources=["genomeview/_quickconsensus.pyx"],
                  include_dirs=get_includes(),
                  define_macros=get_defines()
                  )
          ],
      python_requires=">=3.3"
     )



import itertools

class GraphicsBackend:
    pass
    
def _addOptions(kwdargs, defaults=None):
    if defaults is None: defaults = {}
    options = []
    defaults.update(kwdargs)
    for key, arg in defaults.items():
        if arg is not None and arg != "":
            options.append("""{key}="{arg}" """.format(key=key, arg=arg))
    return "".join(options)

class SVG(GraphicsBackend):
    _filter_id = 0

    def text(self, x, y, text, size=10, anchor="middle", family="Helvetica", **kwdargs):
        defaults = {}
        assert anchor in ["start", "middle", "end"]
        yield """<text x="{x:.2f}" y="{y:.2f}" font-size="{size}" font-family="{family}" text-anchor="{anchor}" {more}>{text}</text>""".format(
            x=x, y=y, size=size, family=family, anchor=anchor, more=_addOptions(kwdargs, defaults), text=text)

    def text_with_background(self, x, y, text, size=10, anchor="middle", text_color="black", bg="white", bg_opacity=0.8, **kwdargs):
        self._filter_id += 1

        text_filter = [
            """<defs>""",
            """    <filter x="0" y="0" width="1" height="1" id="solid{}">""".format(self._filter_id),
            """        <feFlood flood-opacity="{}" flood-color="{}"/>""".format(bg_opacity, bg),
            """        <feComposite in="SourceGraphic"/>""",
            """    </filter>""",
            """</defs>"""]
        for line in text_filter:
            yield line

        # this is a stoopid hack to get the filter to be fully behind the text, without making it blurry
        kwdargs["fill"] = bg
        kwdargs["filter"] = "url(#solid{})".format(self._filter_id)
        yield from self.text(x, y, text, size, anchor, **kwdargs)

        kwdargs["fill"] = text_color
        del kwdargs["filter"]
        yield from self.text(x, y, text, size, anchor, **kwdargs)



    def rect(self, x, y, width, height, **kwdargs):
        defaults = {"fill":"white", "stroke":"black"}
        tag = """<rect x="{x:.2f}" y="{y:.2f}" width="{w:.2f}" height="{h:.2f}" {more}/>""".format(
            x=x, y=y, w=width, h=height, more=_addOptions(kwdargs, defaults))
        yield tag

    def line(self, x1, y1, x2, y2, **kwdargs):
        defaults = {"stroke":"black"}
        yield """<line x1="{x1:.2f}" x2="{x2:.2f}" y1="{y1:.2f}" y2="{y2:.2f}" {more} />""".format(
            x1=x1, x2=x2,  y1=y1, y2=y2, more=_addOptions(kwdargs, defaults))

    def line_with_arrows(self, x1, y1, x2, y2, n=5, direction="right", color="black", arrowKwdArgs=None, **kwdargs):
        self.n = n
        self.direction = direction
        self.arrowKwdArgs = arrowKwdArgs
        
        defaults = {"stroke":color}
        defaults.update(kwdargs)
        
        yield from self.line(x1, y1, x2, y2, **defaults)
        if arrowKwdArgs is None: arrowKwdArgs = {}

        for i in range(1, self.n+1):
            x_arrow = x1+float(x2-x1)*i/n
            y_arrow = y1+float(y2-y1)*i/n
            yield from self.arrow(x_arrow, y_arrow, direction, 
                color=color, scale=arrowKwdArgs.get("stroke-width", 1), **arrowKwdArgs)

    def arrow(self, x, y, direction, color="black", scale=1.0, **kwdargs):
        more = _addOptions(kwdargs)

        if direction == "right":
            a = """<path d="M {x0} {y0} L {x1} {y1} L {x2} {y2} z" fill="{color}" xcenter="{xcenter}" {more}/>""".format(
                x0=(x-10*scale), y0=(y-5*scale), 
                x1=(x), y1=y, 
                x2=(x-10*scale), y2=(y+5*scale),
                color=color,
                xcenter=x,
                more=more)
        elif direction == "left":
            a = """<path d="M {x0} {y0} L {x1} {y1} L {x2} {y2} z" fill="{color}" xcenter="{xcenter}" {more}/>""".format(
                x0=(x+10*scale), y0=(y-5*scale), 
                x1=(x), y1=y, 
                x2=(x+10*scale), y2=(y+5*scale),
                color=color,
                xcenter=x,
                more=more)
        yield a
        
    def start_clipped_group(self, x, y, width, height, name):
        yield """<clipPath id="clip_path_{}"><rect x="{}" y="{}" width="{}" height="{}" /></clipPath>""".format(
            name, x, y, width, height)
        yield """<g clip-path="url(#clip_path_{})">""".format(name)
        
    def stop_clipped_group(self):
        yield "</g>"
    

class Renderer:
    newid = itertools.count()
    
    def __init__(self, backend, x, y, width, height):
        self.backend = backend
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.id = next(Renderer.newid)
    
    def text(self, x, y, *args, **kwdargs):
        yield from self.backend.text(x+self.x, y+self.y, *args, **kwdargs)
        
    def text_with_background(self, x, y, *args, **kwdargs):
        yield from self.backend.text_with_background(x+self.x, y+self.y, *args, **kwdargs)
        
    def rect(self, x, y, *args, **kwdargs):
        yield from self.backend.rect(x+self.x, y+self.y, *args, **kwdargs)
    
    def line(self, x1, y1, x2, y2, *args, **kwdargs):
        yield from self.backend.line(x1+self.x, y1+self.y, x2+self.x, y2+self.y, *args, **kwdargs)
    
    def line_with_arrows(self, x1, y1, x2, y2, *args, **kwdargs):
        yield from self.backend.line_with_arrows(x1+self.x, y1+self.y, x2+self.x, y2+self.y, *args, **kwdargs)

    def arrow(self, x, y, *args, **kwdargs):
        yield from self.backend.arrow(x+self.x, y+self.y, *args, **kwdargs)

    def render(self, element):
        yield "<!-- {} -->".format(element.name)
        yield from self.backend.start_clipped_group(self.x, self.y, self.width, self.height, self.id)
        # yield self.backend.rect(self.x, self.y, self.width, self.height, fill="blue")

        if hasattr(element, "prerenderers"):
            for prerenderer in element.prerenderers:
                for subelement in prerenderer(self, element):
                    yield "   " + subelement

        
        for subelement in element.render(self):
            yield "   " + subelement

        if hasattr(element, "postrenderers"):
            for postrenderer in element.postrenderers:
                for subelement in postrenderer(self, element):
                    yield "   " + subelement

        yield from self.backend.stop_clipped_group()
        
    def subrenderer(self, x=0, y=0, width=None, height=None):
        if width is None: width = self.width
        if height is None: height = self.height

        assert width <= self.width
        assert height <= self.height, "{} {}".format(height, self.height)

        x += self.x
        y += self.y
        
        renderer = Renderer(self.backend, x, y, width, height)

        return renderer



import collections


from genomeview.svg import Renderer, SVG


class Document:
    header = """<svg version="1.1" baseProfile="full" width="{width}" height="{height}" xmlns="http://www.w3.org/2000/svg">"""
    footer = """</svg>"""
        
    def __init__(self, width):
        self.elements = []
        self.width = width
        self.renderer = SVG()
        
        self.margin_x = 5
        self.margin_y = 5

        self.between_views = 5

    def add_view(self, view):
        self.elements.append(view)
        
    def layout(self):
        self.view_width = self.width - self.margin_x*2
        for element in self.elements:
            element.layout(self.view_width)
        
    def render(self):
        self.layout()
        
        total_height = sum(element.height+self.between_views for element in self.elements) + self.margin_y*2
        yield self.header.format(height=total_height, width=self.width)
        
        cury = self.margin_y
        for element in self.elements:
            renderer = Renderer(self.renderer, self.margin_x, cury, self.view_width, element.height)
            yield from renderer.render(element)
            cury += element.height + self.between_views
            
        yield self.footer
        
    def _repr_svg_(self):
        return "\n".join(self.render())
        
class ViewRow:
    def __init__(self, name):
        self.name = name
        self.views = collections.OrderedDict()

        self.width = None
        self.height = None

        self.space_between = 5
    
    def add_view(self, view):
        assert view.name not in self.views
        self.views[view.name] = view
        
    def layout(self, width):
        self.width = width
        n_views = len(self.views)
        self.each_width = (self.width - self.space_between*(n_views-1)) / n_views

        self.height = 0
        for name, view in self.views.items():
            view.layout(self.each_width)
            self.height = max(self.height, view.height)
    
    def render(self, renderer):
        curx = 0
        for name, view in self.views.items():
            subrenderer = renderer.subrenderer(x=curx, width=self.each_width, height=view.height)
            yield from subrenderer.render(view)
            curx += self.each_width + self.space_between
        

class GenomeView:
    def __init__(self, name, chrom, start, end, strand, source):
        self.name = name
        self.tracks = collections.OrderedDict()

        self.scale = Scale(chrom, start, end, strand, source)

        self.pixel_width = None
        self.pixel_height = None

        self.margin_y = 10
    
    def add_track(self, track):
        assert track.name not in self.tracks
        self.tracks[track.name] = track
        
    def layout(self, width):
        self.pixel_width = width
        self.scale.pixel_width = width

        self.height = 0
        for name, track in self.tracks.items():
            track.layout(self.scale)
            self.height += track.height + self.margin_y
    
    def render(self, renderer):
        cury = 0
        for name, track in self.tracks.items():
            subrenderer = renderer.subrenderer(y=cury, height=track.height)
            yield from subrenderer.render(track)
            cury += track.height + self.margin_y
        
    
    
class Scale(object):
    """
    Maintains information about a projection of a specific genomic
    interval into screen coordinates.

    That is, we're interested in visualizing an interval (chrom:start-end) 
    on a canvas of a specified pixel width. The scale enables converting
    genomic coordinates into the display coordinates.
    """
    def __init__(self, chrom, start, end, strand, source):
        self.chrom = chrom
        self.start = start
        self.end = end
        self.strand = strand

        self.pixel_width = None
        self._param = None

        self.source = source
        self._seq = None

    def _setup(self):
        if self._param == self.pixel_width: return
        self._param = self.pixel_width
        self._seq = None

        nt_width = self.end - self.start
        self.bases_per_pixel = nt_width / self.pixel_width

    def topixels(self, genomic_position):
        """
        Converts a genomic position to a pixel location in the current
        coordinate system.
        """
        self._setup()

        pos = (genomic_position - self.start) / float(self.bases_per_pixel)
        return pos

    def relpixels(self, genomic_size):
        """
        Takes a genomic length (ie, a number of basepairs) and converts
        it to a relative screen length in pixels.
        """
        self._setup()

        dist = genomic_size / float(self.bases_per_pixel)
        return dist
    
    def get_seq(self, start=None, end=None, strand="+"):
        """
        Gets the nucleotide sequence of an interval. By default, returns the 
        sequence for the current genomic interval.
        """
        self._setup()

        if start is None:
            start = self.start
        if end is None:
            end = self.end
        
        assert start >= self.start
        assert end <= self.end

        if self._seq is None:
            self._seq = self.source.get_seq(self.chrom, self.start, self.end, self.strand).upper()

        cur_seq = self._seq[start-self.start:end-self.start]
        if strand != self.strand:
            raise Exception("ack")

        return cur_seq

